package nxm.sys.inc;

import java.lang.annotation.Documented;
import java.lang.annotation.Inherited;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;

/** Marker annotation for Java code that has <b>LIMITED FUTURE SUPPORT</b>. Old
    functionality that will have limited future support<br>
    examples:<br>
    old macro-language function<br>
    applets, as Java is dropping support

    @author  Jeanette McCall
    @since   NeXtMidas 3.7.2
 */
@Documented
@Inherited
@Retention(RetentionPolicy.RUNTIME)
//@Target is specifically omitted as omitting it permits all targets
@InternalUseOnly
public @interface LimitedFutureSupport {
  /** Optional description or notes on when, why, etc. marked item has
   * limited future support.
   * @return description or notes on why this tag has limited future support
   * @since 3.7.2
   */
  String value() default "";
}