package nxm.sys.prim;

import nxm.sys.lib.BaseFile;
import nxm.sys.lib.Data;
import nxm.sys.lib.DataFile;
import nxm.sys.lib.DataOp;
import nxm.sys.lib.Parser;
import nxm.sys.lib.Primitive;
import nxm.sys.lib.Sorter;

/**
  Picks multiple peaks from file/pipe

  @author Ken Smith
  @deprecated Since NeXtMidas 3.3.1, use nxm.dsp.prim.peakpick.
 */
@Deprecated
public class peakpick extends Primitive  {

  private DataFile hi, ho_abs, ho_val;
  private Data data_in, data_abs, data_val;
  /** list sorter for sorting Peaks */
  public  Sorter sorter;

  private boolean results_only, init;
  private int framesz, bufsize, maxPeaks, maxInd, inType, outType;
  private float padValue, maxVal, maxAbs, hysteresis;
  private float[] in_fbuf, abs_fbuf, val_fbuf;
  private double start, delta;
  private String out_format, label_val, label_abs;

  //support for complex data added in 3.3.1 (Bug 2660)
  private boolean    complex;      // whether or not the input data is complex
  private int        cxmode;       // mode to use with complex input
  private final int  M_MAG   = 1;  // MAG mode for complex data
  private final int  M_PHASE = 2;  // PHASE mdoe for complex data
  private final int  M_REAL  = 3;  // REAL mode for complex data
  private final int  M_IMAG  = 4;  // IMAG mode for complex data
  private final int  M_10LOG = 5;  // 10LOG mode for complex data
  private final int  M_20LOG = 6;  // 20LOG mode for complex data
  private final static String modeList = "Mag,Phase,Real,Imag,10Log,20Log"; // Supported complex modes

  // A future enhancement would be to use intrinsic Java
  //   classes for sorting, after I get the docs on them.

  public int open () {
    M.deprecate("Since NeXtMidas 3.3.1: use PEAKPICK;DSP instead of PEAKPICK;SYS");

    // Open input file
    hi = MA.getDataFile ("IN");
    hi.open();

    // Error Checking
    if ((hi.typeClass!=1) && (hi.typeClass!=2)) {
      M.error ("Peakpick only handles type 1000/2000 files");
      return(ABORT);
    }

    start = hi.getXStart();
    delta = hi.getXDelta();

    // Get command line parameters

    maxPeaks = MA.getL("MAX_PEAKS");
    padValue = MA.getF("PAD");  // support for PAD added in 3.3.1
    padValue = MA.getF("PAD_VALUE", padValue);  // PAD_VALUE has precedence

    hysteresis = MA.getF("/HYSTER", 0.f);

    // support for complex data added in 3.3.1 (Bug 2660)
    complex    = (hi.getFormatMode()=='C') ? true : false;
    if (complex) {
      String arg = MA.getS("CXMODE", "MAG");  // default to MAGNITUDE
      cxmode = Parser.find(modeList,arg,M_MAG,0,Parser.E_WARNING);
    }

    if (hi.typeClass==1) {
      framesz = (int) hi.getSize();
      inType = 1000;
      outType = 1000;
    } else if (hi.typeClass==2) {
      framesz = hi.getSubSize();
      if (maxPeaks==1) {
        inType = 2000;
        outType = 1000;
      } else {
        inType = 2000;
        outType = 2000;
      }
    }

    // For type 2000 files (or ones we are treating as type 2000),
    //   only read one record at a time (this is set by our buffer size)
    // For type 1000 files, we need to break up the input to separate
    //   buffers to avoid "out of memory" exceptions

    if (inType==2000) {
      bufsize = framesz;
    } else {
      bufsize = 8192;
    }

    if (inType==1000) {
      data_in = hi.getDataBuffer( bufsize, Data.FLOAT );
    } else {
      data_in = hi.getDataBuffer( 1, Data.FLOAT );
    }

    // Open our output files, if appropriate

    if ((inType==1000) && (maxPeaks==1)) {
      // In this case, we write out only result parameters
      results_only = true;
      if (MA.find("OUT_VAL")) {
        label_val = MA.getS("OUT_VAL");
      } else {
        label_val = null;
      }
      if (MA.find("OUT_ABS")) {
        label_abs = MA.getS("OUT_ABS");
      } else {
        label_abs = null;
      }
    } else {
      results_only = false;
      ho_abs = MA.getDataFile("OUT_ABS", hi, 0);
      ho_abs.setDataSize(0.);        // Why do I have to do this????
      ho_abs.setSubSize(maxPeaks);   // This must be before setType
      ho_abs.setType(outType);       //   or it blows up
      ho_abs.setFormat("SF");
      ho_abs.open(BaseFile.OUTPUT|BaseFile.OPTIONAL);

      ho_val = MA.getDataFile("OUT_VAL", hi, 0);
      ho_val.setDataSize(0.);  // Why do I have to do this????
      ho_val.setType(outType);
      ho_val.setFormat("SF");
      ho_val.setSubSize(maxPeaks);
      ho_val.open(BaseFile.OUTPUT|BaseFile.OPTIONAL);

    }
    // Create our data buffers

    if (!results_only) {
      data_abs = new Data("SF", maxPeaks);
      data_val = new Data("SF", maxPeaks);
      abs_fbuf = new float[maxPeaks];
      val_fbuf = new float[maxPeaks];
    }

    if (maxPeaks > 1) sorter = new Sorter(maxPeaks);

    // Initialize our flag, in case we need multiple reads to finish

    init = false;

    maxVal = 0.f;             // These aren't really necessary, but they
    maxInd = 0;               //   keep the stupid compiler happy
    return (NORMAL);
  }

  public int process () {

    boolean clearArrays, peakLogged;
    int i, offset, ngot;
    float value, abs;

    ngot = hi.read(data_in);
    if (ngot==0) return (NOOP);

    // If we never started and have no data, we got an empty file, bail out

    if ( ngot<0 ) {
      if (init && results_only) {
        if (label_abs != null) {
          MR.put(label_abs, maxAbs);
        }
        if (label_val != null) {
          MR.put(label_val, maxVal);
        }
        return(FINISH);
      } else if (!init) {
        return (FINISH);
      } else {
        return(FINISH);
      }
    } else {
      if (!init) {
        clearArrays = true;
      } else {
        clearArrays = false;
      }
      init = true;
    }

    if (inType==2000) {
      //ngot = ngot * data_in.size * hi.getSubSize();   // why was this using the buffer size?
      ngot = ngot * hi.getSubSize();  // changed from line above in 3.3.1 for Bug 2660
    }

    // complex data handling added in 3.3.1 for Bug 2660
    if (complex) {
      switch (cxmode) {
        case  M_MAG   : DataOp.mag     (data_in, data_in, ngot, DataOp.CS); break;
        case  M_PHASE : DataOp.phase   (data_in, data_in, ngot, DataOp.CS); break;
        case  M_REAL  : DataOp.real    (data_in, data_in, ngot, DataOp.CS); break;
        case  M_IMAG  : DataOp.imag    (data_in, data_in, ngot, DataOp.CS); break;
        case  M_10LOG : DataOp.dblogb  (data_in, data_in, ngot, DataOp.CS); break;
        case  M_20LOG : DataOp.dblogb2 (data_in, data_in, ngot, DataOp.CS); break;
        default       : DataOp.mag     (data_in, data_in, ngot, DataOp.CS); break;
      }
    }

    in_fbuf = data_in.castF(true); // Loads data to array

    // Special case if we only want one peak per "frame"
    if (maxPeaks == 1) {
      maxVal = in_fbuf[0];
      maxInd = 1;

      for (int j = 1; j < ngot; j++) {
        if (in_fbuf[j] > maxVal) {
          maxVal = in_fbuf[j];
          maxInd = j;
        }
      }
      maxAbs = (float) start + maxInd * (float) delta;
      if (!results_only) {
        if (ho_val.isOpen) {
          val_fbuf[0] = maxVal;
          data_val.fromArray(val_fbuf);
          ho_val.write(data_val, 1);
        }
        if (ho_abs.isOpen) {
          abs_fbuf[0] = maxAbs;
          data_abs.fromArray(abs_fbuf);
          ho_abs.write(data_abs, 1);
        }
      }

    } else {
      // In this case, we may have multiple output peaks
      //    per input frame so need to keep them sorted.
      // For type 2000 inputs, we need to clear out our
      //   sorter each time.  For type 1000 inputs,
      //   we do it only once.

      if (inType==2000 || clearArrays) {
        sorter.init();
      }
      maxVal = in_fbuf[0];
      maxInd = 1;
      peakLogged = false;

      for (int j = 1; j < ngot; j++) {

        if (in_fbuf[j] > maxVal) {

          // We're going up a peak, log our position

          maxVal = in_fbuf[j];
          maxInd = j;
          peakLogged = false;

        } else if (!peakLogged && (maxVal - in_fbuf[j] > hysteresis) ) {

          // We've gone far enough down to declare a hit, any future peak is separate

          maxAbs = (float) hi.getXStart() + maxInd * (float) hi.getXDelta();
          sorter.add(maxVal, maxAbs);
          peakLogged = true;
          maxVal = in_fbuf[j];

        } else if (peakLogged) {

          // We're just looking for the next peak, make sure our peak
          //   logger doesn't hang up at an old high value.

          maxVal = in_fbuf[j];
        }
      }

      // If our input is type 2000, write our array of outputs

      if (inType == 2000) doOutput();
    }
    //If our input is type 1000 and we have multiple peaks, write our output now

    if (inType == 1000 && (maxPeaks > 1) ) doOutput();
    data_in.uncast(in_fbuf,false); // Loads data to array

    return(NORMAL);
  }

  public int close () {
    hi.close();
    if (!results_only) {  // Unlike X-Midas, closing an unused HCB
      ho_abs.close();   //   causes an error
      ho_val.close();
    }
    return(NORMAL);
  }

  private void doOutput () {

    int numOut;

    for (int j = 0; j < maxPeaks; j++) {
      val_fbuf[j] = sorter.getKey(j, padValue);
      abs_fbuf[j] = sorter.getVal(j, padValue);
    }

    if (outType == 2000) {
      numOut = 1;
    } else {
      numOut = maxPeaks;
    }

    if (ho_abs.isOpen) {
      data_abs.fromArray(abs_fbuf);
      ho_abs.write(data_abs, numOut);
    }
    if (ho_val.isOpen) {
      data_val.fromArray(val_fbuf);
      ho_val.write(data_val, numOut);
    }
  }
}
